# /packages/intranet-invoicing/tcl/intranet-cost-procs.tcl
#
# Copyright (C) 2003-2004 Project/Open
#
# All rights reserved. Please check
# http://www.project-open.com/license/ for details.

ad_library {
    Bring together all "components" (=HTML + SQL code)
    related to Costs

    @author frank.bergann@project-open.com
}

# ---------------------------------------------------------------
# Stati and Types
# ---------------------------------------------------------------

# Frequently used Costs Stati
ad_proc -public im_cost_status_created {} { return 3802 }
ad_proc -public im_cost_status_outstanding {} { return 3804 }
ad_proc -public im_cost_status_past_due {} { return 3806 }
ad_proc -public im_cost_status_partially_paid {} { return 3808 }
ad_proc -public im_cost_status_paid {} { return 3810 }
ad_proc -public im_cost_status_deleted {} { return 3812 }
ad_proc -public im_cost_status_filed {} { return 3814 }
ad_proc -public im_cost_status_requested {} { return 10000090 }
ad_proc -public im_cost_status_rejected {} { return 3818 }



# Frequently used Cost Types
ad_proc -public im_cost_type_invoice {} { return 3700 }
ad_proc -public im_cost_type_quote {} { return 3702 }
ad_proc -public im_cost_type_bill {} { return 3704 }
ad_proc -public im_cost_type_po {} { return 3706 }
ad_proc -public im_cost_type_company_doc {} { return 3708 }
ad_proc -public im_cost_type_provider_doc {} { return 3710 }
ad_proc -public im_cost_type_provider_travel {} { return 3712 }
ad_proc -public im_cost_type_employee {} { return 3714 }
ad_proc -public im_cost_type_repeating {} { return 3716 }
ad_proc -public im_cost_type_timesheet {} { return 3718 }
ad_proc -public im_cost_type_expense_item {} { return 3720 }
ad_proc -public im_cost_type_expense_bundle {} { return 3722 }
ad_proc -public im_cost_type_delivery_note {} { return 3724 }
ad_proc -public im_cost_type_timesheet_planned {} { return 3726 }
ad_proc -public im_cost_type_expense_planned {} { return 3728 }


ad_proc -public im_cost_type_short_name { cost_type_id } { 
    switch $cost_type_id {
	3700 { return "invoice" }
	3702 { return "quote" }
	3704 { return "bill" }
	3706 { return "po" }
	3708 { return "customer_doc" }
	3710 { return "provider_doc" }
	3712 { return "provider_travel" }
	3714 { return "employee" }
	3716 { return "repeating" }
	3718 { return "timesheet" }
	3720 { return "expense" }
	3722 { return "expense_bundle" }
	3724 { return "delivery_note" }
	default { return "unknown" }
    }
}


# Payment Methods
ad_proc -public im_payment_method_undefined {} { return 800 }
ad_proc -public im_payment_method_cash {} { return 802 }


ad_proc -public im_package_cost_id { } {
} {
    return [util_memoize "im_package_cost_id_helper"]
}

ad_proc -private im_package_cost_id_helper {} {
    return [db_string im_package_core_id {
        select package_id from apm_packages
        where package_key = 'intranet-cost'
    } -default 0]
}




# -----------------------------------------------------------
# Characteristics & Grouping
# -----------------------------------------------------------

ad_proc -public im_cost_type_is_invoice_or_quote_p { cost_type_id } {
    Invoices and Quotes have a "Company" fields,
    so we need to identify them:
} {
    set invoice_or_quote_p [expr $cost_type_id == [im_cost_type_invoice] || $cost_type_id == [im_cost_type_quote] || $cost_type_id == [im_cost_type_delivery_note]]
    return $invoice_or_quote_p
}


ad_proc -public im_cost_type_is_invoice_or_bill_p { cost_type_id } {
    Invoices and Bills have a "Payment Terms" field.
    So we need to identify them:
} {
    set invoice_or_bill_p [expr $cost_type_id == [im_cost_type_invoice] || $cost_type_id == [im_cost_type_bill]]
    return $invoice_or_bill_p
}


# -----------------------------------------------------------
# Permissions
# -----------------------------------------------------------

ad_proc -public im_cost_permissions {user_id cost_id view_var read_var write_var admin_var} {
    Fill the "by-reference" variables read, write and admin
    with the permissions of $user_id on $cost_id.<br>

    Cost permissions depend on the rights of the underlying company
    and on Cost Center permissions.
} {
    upvar $view_var view
    upvar $read_var read
    upvar $write_var write
    upvar $admin_var admin

    set user_is_freelance_p [im_user_is_freelance_p $user_id]
    set user_is_customer_p [im_user_is_customer_p $user_id]

    # -----------------------------------------------------
    # Get Cost information
    set customer_id 0
    set provider_id 0
    set cost_center_id 0
    set cost_type_id 0
    db_0or1row get_companies "
        select
                customer_id,
                provider_id,
		cost_center_id,
		cost_type_id
        from
                im_costs
        where
                cost_id = :cost_id
    "

    # -----------------------------------------------------
    # Cost Center permissions - check if the user has read permissions
    # for this particular cost center
    set cc_read [im_cost_center_read_p $cost_center_id $cost_type_id $user_id]
    set cc_write [im_cost_center_write_p $cost_center_id $cost_type_id $user_id]

    set can_read [expr [im_permission $user_id view_costs] || [im_permission $user_id view_invoices]]
    set can_write [expr [im_permission $user_id add_costs] || [im_permission $user_id add_invoices]]

    # AND-connection with add/view - costs/invoices
    if {!$can_read} { set cc_read 0 }
    if {!$can_write} { set cc_write 0 }

    # Set the other two variables
    set cc_admin $cc_write
    set cc_view $cc_read

    # -----------------------------------------------------
    # Customers get the right to see _their_ invoices
    set cust_view 0
    set cust_read 0
    set cust_write 0
    set cust_admin 0
    if {$user_is_customer_p && $customer_id && $customer_id != [im_company_internal]} {
        im_company_permissions $user_id $customer_id cust_view cust_read cust_write cust_admin
    }

    # -----------------------------------------------------
    # Providers get the right to see _their_ invoices
    # This leads to the fact that FreelanceManagers (the guys
    # who can convert themselves into freelancers) can also
    # see the freelancer's permissions. Is this desired?
    # I guess yes, even if they don't usually have the permission
    # to see finance.
    set prov_view 0
    set prov_read 0
    set prov_write 0
    set prov_admin 0
    if {$user_is_freelance_p && $provider_id && $provider_id != [im_company_internal]} {
        im_company_permissions $user_id $provider_id prov_view prov_read prov_write prov_admin
    }

    # -----------------------------------------------------
    # Set the permission as the OR-conjunction of provider and customer
    set view [expr $cust_view || $prov_view || $cc_view]
    set read [expr $cust_read || $prov_read || $cc_read]
    set write [expr $cust_write || $prov_write || $cc_write]
    set admin [expr $cust_admin || $prov_admin || $cc_admin]

    # Limit rights of all users to view & read if they dont
    # have the expressive permission to "add_costs or add_invoices".
    if {!$can_write} {
        set write 0
        set admin 0
    }
}





ad_proc -public im_cost_type_write_p {
    user_id
    cost_type_id
} {
    Returns "1" if the user can create costs of type cost_type_id or 0 otherwise
} {
    set create_cost_types [im_cost_type_write_permissions $user_id]
    return [expr [lsearch -exact $create_cost_types $cost_type_id] != -1]
}


ad_proc -public im_cost_type_write_permissions {
    user_id
} {
    Returns a list of all cost_type_ids for which the user has
    write permissions for atleast one Cost Center.
} {
    return [util_memoize "im_cost_type_write_permissions_helper $user_id" 60]
}


ad_proc -public im_cost_type_write_permissions_helper {
    user_id 
} {
    Returns a list of all cost_type_ids for which the user has
    write permissions for atleast one Cost Center.
} {
    # Financial Write permissions are required
    set can_write [expr [im_permission $user_id add_costs] || [im_permission $user_id add_invoices]]
    if {!$can_write} { return [list] }

    set result [db_list writable_cost_centers "
	select distinct
		ct.cost_type_id
	from
		im_cost_centers cc,
		im_cost_types ct,
	        acs_permissions p,
	        party_approved_member_map m,
	        acs_object_context_index c,
	        acs_privilege_descendant_map h
	where
		cc.cost_center_id = c.object_id
		and ct.write_privilege = h.descendant
	        and p.object_id = c.ancestor_id
	        and m.member_id = :user_id
	        and p.privilege = h.privilege
	        and p.grantee_id = m.party_id
    "]
    return $result
}



# -----------------------------------------------------------
# Options & Selects
# -----------------------------------------------------------

ad_proc -public im_cost_center_status_options { {include_empty 1} } { 
    Cost Center status options
} {
    set options [db_list_of_lists cost_center_status_options "
	select	category, category_id 
	from	im_categories
	where	category_type = 'Intranet Cost Center Status' and
		(enabled_p is null OR enabled_p = 't')
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}


ad_proc -public im_cost_center_type_options { {include_empty 1} } { 
    Cost Center type options
} {
    set options [db_list_of_lists cost_center_type_options "
	select	category, category_id 
	from	im_categories
	where	category_type = 'Intranet Cost Center Type' and
		(enabled_p is null OR enabled_p = 't')
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}


ad_proc -public im_cost_uom_options { {include_empty 1} } {
    Cost UoM (Unit of Measure) options
} {
    set options [db_list_of_lists cost_type_options "
        select	category, category_id
        from	im_categories
	where	category_type = 'Intranet UoM' and
		(enabled_p is null OR enabled_p = 't')
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}



# ---------------------------------------------------------------
# Auxil
# ---------------------------------------------------------------

ad_proc -public n20 { value } { 
    Converts null ("") values to numeric "0". This function
    is used inside view definitions in order to deal with null
    values in TCL 
} {
    if {"" == $value} { set value 0 }
    return $value
}



# ---------------------------------------------------------------
# Cost Item Creation
# ---------------------------------------------------------------

namespace eval im_cost {

    ad_proc -public new { 
	{-cost_id 0}
	{-cost_status_id 0}
	{-customer_id 0}
	{-provider_id 0}
	{-user_id 0}
	{-creation_ip ""}
	{-object_type "im_cost"}
	-cost_name
	-cost_type_id
    } {
	Create a new cost item and return its ID.
	The cost item is created with the minimum number
	of parameters necessary for creation (not null).
	We asume that an UPDATE is executed afterwards
	in order to fill in more details.
    } {
	if {0 == $customer_id} { set customer_id [im_company_internal] }
	if {0 == $provider_id} { set provider_id [im_company_internal] }
	if {0 == $cost_id} { set cost_id [db_nextval acs_object_id_seq]}
	if {0 == $cost_status_id} { set cost_status_id [im_cost_status_created]}
	if {0 == $user_id} { set user_id [ad_get_user_id] }
	if {"" == $creation_ip} { set creation_ip [ad_conn peeraddr] }

	set today [db_string today "select sysdate from dual"]
	if [catch {
	    db_dml insert_costs "
		insert into acs_objects	(
			object_id, object_type, context_id,
			creation_date, creation_user, creation_ip
		) values (
			:cost_id, :object_type, null,
			:today, :user_id, :creation_ip
	    )" 
	} errmsg] {
	    ad_return_complaint 1 "<li>Error creating acs_object in im_cost::new<br>
		cost_id=$cost_id<br>
		cost_status_id=$cost_status_id<br>
		customer_id=$customer_id<br>
		provider_id=$provider_id<br>
		cost_name=$cost_name<br>
		cost_type_id=$cost_type_id<br>
            <pre>$errmsg</pre>"
	    return
	}

	if [catch {
	    db_dml insert_costs "
	insert into im_costs (
		cost_id,
		cost_nr,
		cost_name,
		customer_id,
		provider_id,
		cost_type_id,
		cost_status_id
	) values (
		:cost_id,
		:cost_id,
		:cost_name,
		:customer_id,
		:provider_id,
		:cost_type_id,
		:cost_status_id
	)" } errmsg] {
	    ad_return_complaint 1 "<li>Error inserting into im_costs as part of im_cost::new<br>
               cost_id=$cost_id<br>
                cost_status_id=$cost_status_id<br>
                customer_id=$customer_id<br>
                provider_id=$provider_id<br>
                cost_name=$cost_name<br>
                cost_type_id=$cost_type_id<br>
            <pre>$errmsg</pre>"
	    return
	}
	return $cost_id
    }
    
}

# ---------------------------------------------------------------
# Options for Form elements
# ---------------------------------------------------------------

ad_proc -public im_cost_type_options { {include_empty 1} } { 
    Cost type options
} {
   set options [db_list_of_lists cost_type_options "
	select cost_type, cost_type_id
	from im_cost_types
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}

ad_proc -public im_cost_status_options { {include_empty 1} } { 
    Cost status options
} {
    set options [db_list_of_lists cost_status_options "
	select cost_status, cost_status_id from im_cost_status
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}

ad_proc -public im_cost_template_options { {include_empty 1} } { 
    Cost Template options
} {
    set options [db_list_of_lists template_options "
	select category, category_id
	from im_categories
	where category_type = 'Intranet Cost Template' and
		(enabled_p is null OR enabled_p = 't')
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}

ad_proc -public im_investment_options { {include_empty 1} } { 
    Cost investment options
} {
    set options [db_list_of_lists investment_options "
	select name, investment_id
	from im_investments
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}

ad_proc -public im_currency_options { {include_empty 1} } { 
    Cost currency options
} {
    set options [db_list_of_lists currency_options "
	select	iso, iso
	from	currency_codes
	where	supported_p = 't'
	order by iso
    "]
    if {$include_empty} { set options [linsert $options 0 { "" "" }] }
    return $options
}

ad_proc im_currency_select {
    {-translate_p 0}
    {-include_empty_p 1}
    {-include_empty_name ""}
    select_name 
    {default ""}
} {
    Return a HTML widget that selects a currency code from
    the list of global countries.
} {
    set bind_vars [ns_set create]
    set statement_name "currency_code_select"
    set sql "select iso, iso
	     from currency_codes
	     where supported_p='t'
	     order by lower(currency_name)"

    return [im_selection_to_select_box -translate_p 0 -include_empty_p $include_empty_p -include_empty_name $include_empty_name $bind_vars $statement_name $sql $select_name $default]
}

ad_proc im_supported_currencies { } {
    Returns the list of supported currencies
} {
    set include_empty 0
    set currency_options [im_currency_options $include_empty]
    set result [list]
    foreach option $currency_options {
	lappend result [lindex $option 0]
    }
    return $result
}



ad_proc -public im_department_options { {include_empty 0} } {
    Returns a list of all Departments in the company.
} {
    set department_only_p 1
    return [im_cost_center_options -include_empty $include_empty -department_only_p $department_only_p]
}

ad_proc -public im_cost_center_select { 
    {-include_empty 0} 
    {-department_only_p 0} 
    select_name 
    {default ""} 
    {cost_type_id ""} 
} {
    Returns a select box with all Cost Centers in the company.
} {
    set options [im_cost_center_options -include_empty $include_empty -department_only_p $department_only_p -cost_type_id $cost_type_id]

    # Only one option, so 
    # write out string instead of select component
    if {[llength $options] == 1} {
	set cc_entry [lindex $options 0]
	set cc_id [lindex $cc_entry 1]
	set cc_name [string trim [lindex $cc_entry 0]]
	# Replace &nbsp; by " "
	regsub -all {\&nbsp\;} $cc_name " " cc_name
	return "$cc_name <input type=hidden name=\"$select_name\" value=\"$cc_id\">\n"
    }

    return [im_options_to_select_box $select_name $options $default]
}


ad_proc -public im_cost_center_options { 
    {-include_empty 0} 
    {-department_only_p 0} 
    {-cost_type_id ""} 
} {
    Returns a list of all Cost Centers in the company.
    Takes into account the permission of the user to 
    charge FinDocs to CostCenters
} {
    set user_id [ad_get_user_id]
    set start_center_id [im_cost_center_company]
    set cost_type "Invalid"
    if {"" != $cost_type_id} { set cost_type [db_string ct "select im_category_from_id(:cost_type_id)"] }

    set cost_type_sql ""
    set short_name [im_cost_type_short_name $cost_type_id]

    # Profit-Center-Module (perms on CCs) installed?
    set pcenter_p [util_memoize "db_string pcent {select count(*) from apm_packages where package_key = 'intranet-cost-center'}"]

    if {$pcenter_p && "" != $cost_type_id} { 
	set cost_type_sql "and im_object_permission_p(cost_center_id, :user_id, 'fi_write_${short_name}s') = 't'\n"
    }

    set department_only_sql ""
    if {$department_only_p} {
	set department_only_sql "and cc.department_p = 't'"
    }

    set options_sql "
        select	cc.cost_center_name,
                cc.cost_center_id,
                cc.cost_center_label,
    		(length(cc.cost_center_code) / 2) - 1 as indent_level
        from	im_cost_centers cc
	where	1=1
		$department_only_sql
		$cost_type_sql
	order by
		cc.cost_center_code
    "

    set options [list]
    if {$include_empty} { lappend options [list "" ""] }

    db_foreach cost_center_options $options_sql {
        set spaces ""
        for {set i 0} {$i < $indent_level} { incr i } {
            append spaces "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
        }
        lappend options [list "$spaces$cost_center_name" $cost_center_id]
    }

    if {$include_empty && [llength $options] == 0} {
	set invalid_cc [lang::message::lookup "" intranet-cost.No_CC_permissions_for_cost_type "No CC permissions for \"%cost_type%\""]
	lappend options [list $invalid_cc ""]
    }

    return $options
}


ad_proc -public template::widget::im_currencies { element_reference tag_attributes } {
    ad_form widget for active currencies
    The widget displays a list of active currencies.
} {
    # Defaults
    set include_empty_p 1
   
    # Get references to parameters (magic...)
    upvar $element_reference element
    array set attributes $tag_attributes
    set field_name $element(name)
    set default_value_list $element(values)

    # Determine parameters
    if { [info exists element(custom)] } {
	set params $element(custom)

	set include_empty_pos [lsearch $params include_empty_p]
	if { $include_empty_pos >= 0 } {
	    set include_empty_p [lindex $params [expr $include_empty_pos + 1]]
	}
    }

    # Determine the default value for the widget
    set default_value ""
    if {[info exists element(value)]} {
	set default_value $element(values)
    }

    # Render the widget, depending on the display_mode (edit/display):
    if { "edit" == $element(mode)} {

	return [im_currency_select $field_name $default_value]

    } else {

	return $default_value
    }
}



ad_proc -public template::widget::im_cost_center_tree { element_reference tag_attributes } {
    ad_form tree widget for cost centers and departments.
    <tt>Usage: {custom {department_only_p 1} {start_cc_id 1234} {include_empty_p 0}}</tt>

    @param start_cc_id
	Set to a cc_id in order to display only a subtree

    @param department_only_p 
	Set to 1 in order to show only departments (subset of CCs)

    @param include_empty_p
	Set to 1 in order to include an empty line at the top

    @param translate_p
	Set to 1 in order to use L10n

    The widget displays a department tree, starting at the top (or start_cc_id).
} {
    # Defaults
    set include_empty_p 1
    set start_cc_id ""
    set department_only_p 1
    set tranlate_p 0
   
    # Get references to parameters (magic...)
    upvar $element_reference element
    array set attributes $tag_attributes
    set field_name $element(name)
    set default_value_list $element(values)

    # Determine parameters
    if { [info exists element(custom)] } {
	set params $element(custom)

	set include_empty_pos [lsearch $params include_empty_p]
	if { $include_empty_pos >= 0 } {
	    set include_empty_p [lindex $params [expr $include_empty_pos + 1]]
	}
	set department_only_p_pos [lsearch $params department_only_p]
	if { $department_only_p_pos >= 0 } {
	    set department_only_p [lindex $params [expr $department_only_p_pos + 1]]
	}
	set translate_p_pos [lsearch $params translate_p]
	if { $translate_p_pos >= 0 } {
	    set translate_p [lindex $params [expr $translate_p_pos + 1]]
	}
    }

    # Determine the default value for the widget
    set default_value ""
    if {[info exists element(value)]} {
	set default_value $element(values)
    }

    # Debug
    if {0} {
	set debug ""
	foreach key [array names element] {
	    set value $element($key)
	    append debug "$key = $value\n"
	}
	ad_return_complaint 1 "<pre>$element(name)\n$debug\n</pre>"
	return
    }

    # Render the widget, depending on the display_mode (edit/display):
    if { "edit" == $element(mode)} {

	return [im_cost_center_select \
		    -include_empty $include_empty_p \
		    -department_only_p $department_only_p \
		    $field_name \
		    $default_value \
	]

    } else {

	if {"" != $default_value && "\{\}" != $default_value} {
	    return [db_string cat "select im_category_from_id($default_value)"]
	}
	return ""
    }
}


ad_proc -public im_costs_default_cost_center_for_user { 
    user_id
} {
    Returns a reasonable default cost center for a given user.
} {
    # For an employee return the department_id:
    set cost_center_id [db_string employee_cost_center "
	select	department_id
	from	im_employees
	where	employee_id = :user_id
    " -default 0]

    if {"" == $cost_center_id || 0 == $cost_center_id} {
	set cost_center_id [im_cost_center_company]
    }

    return $cost_center_id
}


ad_proc -public im_cost_center_company {
} {
    Returns the ID of the "Company - Co" Cost Center
} {
    set cc [db_string cost_center_company "
	select	cost_center_id
	from	im_cost_centers
	where	cost_center_label = 'company'
    " -default 0]

    if {0 == $cc} { 
	set cc [db_string cost_center_company "
		select	cost_center_id
		from	im_cost_centers
		order by length(cost_center_code)
		LIMIT 1
	" -default 0]
    }

    if {0 == $cc} { ad_return_complaint 1 "Unable to find Cost Center 'company'" }
    return $cc
}

ad_proc -public -deprecated im_cost_center_read_p {
    cost_center_id
    cost_type_id
    user_id
} {
    Returns "1" if the user can read the CC, "0" otherwise.
    This TCL-level query makes sense, because it is cached
    and thus quite cheap to execute, while executing the
    acs_permission__permission_p() query could be quite
    expensive with a considerable number of financial docs.
} {
    if {"" == $cost_center_id} { return 1 }
    return [string equal "t" [util_memoize "im_cost_center_read_p_helper $cost_center_id $cost_type_id $user_id" 60]]
}

ad_proc -private -deprecated im_cost_center_read_p_helper {
    cost_center_id
    cost_type_id
    user_id
} {
    Returns "t" if the user can read the CC, "f" otherwise.
} {
    # User can read all CCs if no Profit Center Controlling is installed
    set pcenter_p [util_memoize "db_string pcent {select count(*) from apm_packages where package_key = 'intranet-cost-center'}"]
    if {!$pcenter_p} { return "t" }

    return [db_string cc_perms "
	select	im_object_permission_p(:cost_center_id, :user_id, ct.read_privilege)
	from	im_cost_types ct
	where	ct.cost_type_id = :cost_type_id
    " -default "f"]
}


ad_proc -public im_cc_read_p {
    {-user_id 0}
    {-cost_center_id 0}
    {-cost_type_id 0}
    {-privilege ""}
} {
    Returns "1" if the user can read the global "company" CC
} {
    # User can read all CCs if no Profit Center Controlling is installed
    set pcenter_p [util_memoize "db_string pcent {select count(*) from apm_packages where package_key = 'intranet-cost-center'}"]
    if {!$pcenter_p} { return 1 }

    # Deal with exceptions
    if {0 == $user_id} { set user_id [ad_get_user_id] }
    if {0 == $cost_center_id} { set cost_center_id [im_cost_center_company] }
    if {0 != $cost_type_id} {
	set privilege [util_memoize "db_string priv \"
		select read_privilege 
		from im_cost_types 
		where cost_type_id = $cost_type_id
	\" -default \"\" "]
    }
    if {"" == $privilege} { set privilege "fi_read_all" }

    set true_false [util_memoize "db_string company_cc_read \"
	select	im_object_permission_p($cost_center_id, $user_id, '$privilege')
    \" -default f" 60]
    return [string equal "t" $true_false]
}

ad_proc -public im_cost_center_write_p {
    cost_center_id
    cost_type_id
    user_id
} {
    Returns "1" if the user can write to the CC, "0" otherwise.
    This TCL-level query makes sense, because it is cached
    and thus quite cheap to execute, while executing the
    acs_permission__permission_p() query could be quite
    expensive with a considerable number of financial docs.
} {
    if {"" == $cost_center_id} { return 1 }
    return [string equal "t" [util_memoize "im_cost_center_write_p_helper $cost_center_id $cost_type_id $user_id" 60]]
}

ad_proc -public im_cost_center_write_p_helper {
    cost_center_id
    cost_type_id
    user_id
} {
    Returns "t" if the user can write to the CC, "f" otherwise.
} {
    # User can write all CCs if no Profit Center Controlling is installed
    set pcenter_p [util_memoize "db_string pcent {select count(*) from apm_packages where package_key = 'intranet-cost-center'}"]
    if {!$pcenter_p} { return "t" }

    return [db_string cc_perms "
	select	im_object_permission_p(:cost_center_id, :user_id, ct.write_privilege)
	from	im_cost_types ct
	where	ct.cost_type_id = :cost_type_id
    " -default "f"]
}


ad_proc -public im_costs_navbar { default_letter base_url next_page_url prev_page_url export_var_list {select_label ""} } {
    Returns rendered HTML code for a horizontal sub-navigation
    bar for /intranet-cost/.
    The lower part of the navbar also includes an Alpha bar.<br>
    Default_letter==none marks a special behavious, printing no alpha-bar.
} {
    # -------- Defaults -----------------------------
    set user_id [ad_get_user_id]
    set url_stub [ns_urldecode [im_url_with_query]]
    ns_log Notice "im_costs_navbar: url_stub=$url_stub"

    set sel "<td class=tabsel>"
    set nosel "<td class=tabnotsel>"
    set a_white "<a class=whitelink"
    set tdsp "<td>&nbsp;</td>"

    # -------- Calculate Alpha Bar with Pass-Through params -------
    set bind_vars [ns_set create]
    foreach var $export_var_list {
        upvar 1 $var value
        if { [info exists value] } {
            ns_set put $bind_vars $var $value
            ns_log Notice "im_costs_navbar: $var <- $value"
        }
    }
    set alpha_bar [im_alpha_bar -prev_page_url $prev_page_url -next_page_url $next_page_url $base_url $default_letter $bind_vars]

    # Get the Subnavbar
    set parent_menu_sql "select menu_id from im_menus where label='finance'"
    set parent_menu_id [db_string parent_admin_menu $parent_menu_sql -default 0]
    set navbar [im_sub_navbar $parent_menu_id "" $alpha_bar "tabnotsel" $select_label]

    return $navbar
}

ad_proc im_next_cost_nr { } {
    Returns the next free cost number

    Cost_nr's look like: 2003_07_123 with the first 4 digits being
    the current year, the next 2 digits the month and the last 3 digits 
    as the current number  within the month.
    Returns "" if there was an error calculating the number.

    The SQL query works by building the maximum of all numeric (the 8 
    substr comparisons of the last 4 digits) cost numbers
    of the current year/month, adding "+1", and contatenating again with 
    the current year/month.

    This procedure has to deal with the case that
    two user are costs projects concurrently. In this case there may
    be a "raise condition", that two costs are created at the same
    moment. This is possible, because we take the cost numbers from
    im_costs_ACTIVE, which excludes costs in the process of
    generation.
    To deal with this situation, the calling procedure has to double check
    before confirming the cost.
} {
    set sql "
select
	to_char(sysdate, 'YYYY_MM')||'_'||
	trim(to_char(1+max(i.nr),'0000')) as cost_nr
from
	(select substr(cost_nr,9,4) as nr from im_costs
	 where substr(cost_nr, 1,7)=to_char(sysdate, 'YYYY_MM')
	 UNION 
	 select '0000' as nr from dual
	) i
where
        ascii(substr(i.nr,1,1)) > 47 and
        ascii(substr(i.nr,1,1)) < 58 and
        ascii(substr(i.nr,2,1)) > 47 and
        ascii(substr(i.nr,2,1)) < 58 and
        ascii(substr(i.nr,3,1)) > 47 and
        ascii(substr(i.nr,3,1)) < 58 and
        ascii(substr(i.nr,4,1)) > 47 and
        ascii(substr(i.nr,4,1)) < 58
"
    set cost_nr [db_string next_cost_nr $sql -default ""]
    ns_log Notice "im_next_cost_nr: cost_nr=$cost_nr"

    return $cost_nr
}



# ---------------------------------------------------------------
# Components
# ---------------------------------------------------------------

ad_proc im_costs_object_list_component { user_id cost_id return_url } {
    Returns a HTML table containing a list of objects
    associated with a particular financial document.
} {
    set bgcolor(0) "class=roweven"
    set bgcolor(1) "class=rowodd"

    set object_list_sql "
	select distinct
	   	o.object_id,
		acs_object.name(o.object_id) as object_name,
		u.url
	from
	        acs_objects o,
	        acs_rels r,
		im_biz_object_urls u
	where
	        r.object_id_one = o.object_id
	        and r.object_id_two = :cost_id
		and u.object_type = o.object_type
		and u.url_type = 'view'
    "

    set ctr 0
    set object_list_html ""
    db_foreach object_list $object_list_sql {
	append object_list_html "
        <tr $bgcolor([expr $ctr % 2])>
          <td>
            <A href=\"$url$object_id\">$object_name</A>
          </td>
          <td>
            <input type=checkbox name=object_ids.$object_id>
          </td>
        </tr>\n"
	incr ctr
    }

    if {0 == $ctr} {
	append object_list_html "
        <tr $bgcolor([expr $ctr % 2])>
          <td><i>No objects found</i></td>
        </tr>\n"
    }

    return "
      <form action=cost-association-action method=post>
      [export_form_vars cost_id return_url]
      <table border=0 cellspacing=1 cellpadding=1>
        <tr>
          <td align=middle class=rowtitle colspan=2>Related Projects</td>
        </tr>
        $object_list_html
        <tr>
          <td align=right>
            <input type=submit name=add_project_action value='Add a Project'>
            </A>
          </td>
          <td>
            <input type=submit name=del_action value='Del'>
          </td>
        </tr>
      </table>
      </form>
    "
}

ad_proc im_costs_company_component { user_id company_id } {
    Returns a HTML table containing a list of costs for a particular
    company.
} {
    return [im_costs_base_component $user_id $company_id ""]
}

ad_proc im_costs_project_component { user_id project_id } {
    Returns a HTML table containing a list of costs for a particular
    particular project.
} {
    return [im_costs_base_component $user_id "" $project_id]
}


ad_proc im_costs_base_component { user_id {company_id ""} {project_id ""} } {
    Returns a HTML table containing a list of costs for a particular
    company or project.
} {
    if {![im_permission $user_id view_costs]} {
	return ""
    }

    set bgcolor(0) " class=roweven "
    set bgcolor(1) " class=rowodd "
    set max_costs 5
    set colspan 5
    set org_project_id $project_id
    set org_company_id $company_id

    # Where to link when clicking on an object linke? "edit" or "view"?
    set view_mode "view"

    # ----------------- Compose SQL Query --------------------------------
  
    set extra_where [list]
    set extra_from [list]
    set extra_select [list]
    set object_name ""
    set new_doc_args ""
    if {"" != $company_id} { 
	lappend extra_where "(ci.customer_id = :company_id OR ci.provider_id = :company_id)" 
	set object_name [db_string object_name "select company_name from im_companies where company_id = :company_id"]
	set new_doc_args "?company_id=$company_id"
    }

    if {"" != $project_id} { 
	# Select the costs explicitely associated with a project.
	lappend extra_where "
	ci.cost_id in (
		select distinct cost_id 
		from im_costs 
		where project_id=:project_id
	    UNION
		select distinct object_id_two as cost_id
		from acs_rels
		where object_id_one = :project_id
	)" 
	set object_name [db_string object_name "select project_name from im_projects where project_id = :project_id"]
	set new_doc_args "?project_id=$project_id"
    }

    set extra_where_clause [join $extra_where "\n\tand "]
    if {"" != $extra_where_clause} { set extra_where_clause "\n\tand $extra_where_clause" }
    set extra_from_clause [join $extra_from ",\n\t"]
    if {"" != $extra_from_clause} { set extra_from_clause ",\n\t$extra_from_clause" }
    set extra_select_clause [join $extra_select ",\n\t"]
    if {"" != $extra_select_clause} { set extra_select_clause ",\n\t$extra_select_clause" }

    set costs_sql "
	select
		ci.*,
		ci.paid_amount as payment_amount,
		ci.paid_currency as payment_currency,
		url.url,
	        im_category_from_id(ci.cost_status_id) as cost_status,
	        im_category_from_id(ci.cost_type_id) as cost_type,
		to_date(to_char(ci.effective_date,'yyyymmdd'),'yyyymmdd') 
			+ ci.payment_days as calculated_due_date
		$extra_select_clause
	from
		im_costs ci,
		acs_objects o,
	        (select * from im_biz_object_urls where url_type=:view_mode) url,
		(	select distinct
				cc.cost_center_id,
				ct.cost_type_id
			from	im_cost_centers cc,
				im_cost_types ct,
				acs_permissions p,
				party_approved_member_map m,
				acs_object_context_index c, 
				acs_privilege_descendant_map h
			where
				p.object_id = c.ancestor_id
				and h.descendant = ct.read_privilege
				and c.object_id = cc.cost_center_id
				and m.member_id = :user_id
				and p.privilege = h.privilege
				and p.grantee_id = m.party_id
		) readable_ccs
		$extra_from_clause
	where
		ci.cost_id = o.object_id
		and o.object_type = url.object_type
		and ci.cost_center_id = readable_ccs.cost_center_id
		and ci.cost_type_id = readable_ccs.cost_type_id
		$extra_where_clause
	order by
		ci.effective_date desc
    "

    set cost_html "
<table border=0>
  <tr>
    <td colspan=$colspan class=rowtitle align=center>
      [_ intranet-cost.Financial_Documents]
    </td>
  </tr>
  <tr class=rowtitle>
    <td align=center class=rowtitle>[_ intranet-cost.Document]</td>
    <td align=center class=rowtitle>[_ intranet-cost.Type]</td>
    <td align=center class=rowtitle>[_ intranet-cost.Due]</td>
    <td align=center class=rowtitle>[_ intranet-cost.Amount]</td>
    <td align=center class=rowtitle>[_ intranet-cost.Paid]</td>
  </tr>
"
    set ctr 1
    set payment_amount ""
    set payment_currency ""

    db_foreach recent_costs $costs_sql {
	append cost_html "
<tr$bgcolor([expr $ctr % 2])>
  <td><A href=\"$url$cost_id\">[string range $cost_name 0 20]</A></td>
  <td>$cost_type</td>
  <td>$calculated_due_date</td>
  <td>$amount $currency</td>
  <td>$payment_amount $payment_currency</td>
</tr>\n"
	incr ctr
	if {$ctr > $max_costs} { break }
    }

    # Restore the original values after SQL selects
    set project_id $org_project_id
    set company_id $org_company_id

	append cost_html "
<tr$bgcolor([expr $ctr % 2])>
  <td colspan=$colspan>
    <A HREF=/intranet-cost/list?[export_url_vars status_id company_id project_id]>
      [_ intranet-cost.more_costs]
    </A>
  </td>
</tr>\n"

    # Add a reasonable message if there are no documents
    if {$ctr == 1} {
	append cost_html "
<tr$bgcolor([expr $ctr % 2])>
  <td colspan=$colspan align=center>
    <I>[_ intranet-cost.lt_No_financial_document]</I>
  </td>
</tr>\n"
	incr ctr
    }


    # Add some links to create new financial documents
    # if the intranet-invoices module is installed
    if {[db_table_exists im_invoices]} {

	# Project Documents:
	if {"" != $project_id} {

	    append cost_html "
	<tr class=rowplain>
	  <td colspan=$colspan>\n"


	    # Customer invoices: customer = Project Customer, provider = Internal
	    set customer_id [db_string project_customer "select company_id from im_projects where project_id=:project_id" -default ""]
	    set provider_id [im_company_internal]
	    set bind_vars [ad_tcl_vars_to_ns_set customer_id provider_id project_id]
      	    append cost_html [im_menu_ul_list "invoices_customers" $bind_vars]

	    # Provider invoices: customer = Internal, no provider yet defined
	    set customer_id [im_company_internal]
	    set bind_vars [ad_tcl_vars_to_ns_set customer_id project_id]
	    append cost_html [im_menu_ul_list "invoices_providers" $bind_vars]

	    append cost_html "	
	  </td>
	</tr>
	"
	    incr ctr

	} 
    }

    append cost_html "</table>\n"
    return $cost_html
}


# ---------------------------------------------------------------
# Benefits & Loss Calculation per Project
# ---------------------------------------------------------------

ad_proc im_costs_project_finance_component { 
    {-show_details_p 1}
    {-show_summary_p 1}
    {-show_admin_links_p 0}
    user_id 
    project_id 
} {
    Returns a HTML table containing a detailed summary of all
    financial activities of the project. <p>

    The complexity of this component comes from several sources:
    <ul>
    <li>We need to sum up the invoices and sort them into several
        "buckets" that correspond to the different cost types
        such as "customer invoices", "provider purchase orders",
        internal "timesheet costs" etc.
    <li>We can have costs and financial documents (see doc.) in
        several currencies, so we can't just add these together.
        Instead, we need to maintain separate sums per cost type
        and currency.
        Also, costs may have NULL cost values (timesheet costs
        from persons whithout the "hourly cost" defined).
    </ul>

} {
    if {![im_permission $user_id view_costs]} {	return "" }
    if {$show_details_p} { set show_admin_links_p 1 }

    set bgcolor(0) " class=roweven "
    set bgcolor(1) " class=rowodd "
    set colspan 7
    set date_format "YYYY-MM-DD"
    set num_format "9999999999.99"

    set return_url [im_url_with_query]

    # Where to link when clicking on an object link? "edit" or "view"?
    set view_mode "view"

    # Default Currency
    set default_currency [ad_parameter -package_id [im_package_cost_id] "DefaultCurrency" "" "EUR"]

    # project_id may get overwritten by SQL query
    set org_project_id $project_id

    # Get a hash array of subtotals per cost_type
    array set subtotals [im_cost_update_project_cost_cache $project_id]

    # ----------------- Compose Main SQL Query --------------------------------

    set project_cost_ids_sql "
		                select distinct cost_id
		                from im_costs
		                where project_id in (
					select	children.project_id
					from	im_projects parent,
						im_projects children
					where	children.tree_sortkey 
							between parent.tree_sortkey 
							and tree_right(parent.tree_sortkey)
						and parent.project_id = :project_id
				)
			    UNION
				select distinct object_id_two as cost_id
				from acs_rels
				where object_id_one in (
					select	children.project_id
					from	im_projects parent,
						im_projects children
					where	children.tree_sortkey 
							between parent.tree_sortkey 
							and tree_right(parent.tree_sortkey)
						and parent.project_id = :project_id
				)
    "

    set costs_sql "
	select
		ci.*,
		to_char(ci.paid_amount, :num_format) as payment_amount,
		ci.paid_currency as payment_currency,
		to_char(ci.amount, :num_format) as amount,
		to_char(ci.amount * im_exchange_rate(ci.effective_date::date, ci.currency, :default_currency), :num_format) as amount_converted,
		p.project_nr,
		p.project_name,
		cust.company_name as customer_name,
		prov.company_name as provider_name,
		url.url,
		im_category_from_id(ci.cost_status_id) as cost_status,
		im_category_from_id(ci.cost_type_id) as cost_type,
		im_cost_center_code_from_id(ci.cost_center_id) as cost_center_code,
		to_date(to_char(ci.effective_date,:date_format),:date_format) + ci.payment_days as calculated_due_date
	from
		im_costs ci
			LEFT OUTER JOIN im_projects p ON (ci.project_id = p.project_id)
			LEFT OUTER JOIN im_companies cust on (ci.customer_id = cust.company_id)
			LEFT OUTER JOIN im_companies prov on (ci.provider_id = prov.company_id),
		acs_objects o,
		(select * from im_biz_object_urls where url_type=:view_mode) url
	where
		ci.cost_id = o.object_id
		and o.object_type = url.object_type
		and ci.cost_id in (
			$project_cost_ids_sql
		)
	      	and ci.cost_type_id not in (
	                [im_cost_type_employee],
	                [im_cost_type_repeating],
	                [im_cost_type_expense_item]
	        )
	order by
		ci.cost_type_id,
		ci.effective_date desc
    "


    set cost_html "
	<table border=0>
	  <tr>
	    <td colspan=$colspan class=rowtitle align=center>
	      [_ intranet-cost.Financial_Documents]
	    </td>
	  </tr>
	  <tr class=rowtitle>
	    <td align=center class=rowtitle>[_ intranet-cost.Document]</td>
	    <td align=center class=rowtitle>[lang::message::lookup "" intranet-cost.CostCenter_short "CC"]</td>
	    <td align=center class=rowtitle>[_ intranet-cost.Company]</td>
	    <td align=center class=rowtitle>[_ intranet-cost.Due]</td>
	    <td align=center class=rowtitle>[_ intranet-cost.Amount]</td>
	    <td align=center class=rowtitle>[lang::message::lookup "" intranet-cost.Org_Amount "Org"]</td>
	    <td align=center class=rowtitle>[_ intranet-cost.Paid]</td>
	  </tr>
    "

    set can_read_summary_p 1

    set ctr 1
    set atleast_one_unreadable_p 0
    set old_atleast_one_unreadable_p 0
    set payment_amount ""
    set payment_currency ""

    set old_project_nr ""
    set old_cost_type_id 0
    db_foreach recent_costs $costs_sql {

	# Write the subtotal line of the last cost_type_id section
	if {$cost_type_id != $old_cost_type_id} {
	    if {0 != $old_cost_type_id} {
		if {!$atleast_one_unreadable_p} {
		    append cost_html "
			<tr class=rowplain>
			  <td colspan=[expr $colspan-3]>&nbsp;</td>
			  <td colspan=2>
			    <b>$subtotals($old_cost_type_id) $default_currency</b>
			  </td>
			</tr>
		    "
		}
		append cost_html "
		<tr>
		  <td colspan=99 class=rowplain>&nbsp;</td>
		</tr>\n"
	    }

	    append cost_html "
		<tr class=rowtitle>
		  <td class=rowplain colspan=99>$cost_type</td>
		</tr>\n"

	    set old_cost_type_id $cost_type_id
	    set old_atleast_one_unreadable_p $atleast_one_unreadable_p
	    set atleast_one_unreadable_p 0
	}

	# Check permissions - query is cached
	set read_p [im_cost_center_read_p $cost_center_id $cost_type_id $user_id]
	if {!$read_p} { 
	    set atleast_one_unreadable_p 1 
	    set can_read_summary_p 0
	}

	set company_name ""
	if {$cost_type_id == [im_cost_type_invoice] || $cost_type_id == [im_cost_type_quote] || $cost_type_id == [im_cost_type_delivery_note]} {
	    set company_name $customer_name
	} else {
	    set company_name $provider_name
	}
	
	set cost_url "<A title=\"$cost_name\" href=\"$url$cost_id&return_url=[ns_urlencode $return_url]\">"
	set cost_url_end "</A>"

	set amount_unconverted "<nobr>([string trim $amount] $currency)</nobr>"
	if {[string equal $currency $default_currency]} { set amount_unconverted "" }

	set amount_paid "$payment_amount $default_currency"
	if {"" == $payment_amount} { set amount_paid "" }

	set default_currency_read_p $default_currency
	if {!$read_p} {
	    set cost_url ""
	    set cost_url_end ""
	    set amount_converted ""
	    set amount_unconverted ""
	    set amount_paid ""
	    set default_currency_read_p ""
	}

	append cost_html "
	<tr $bgcolor([expr $ctr % 2])>
	  <td><nobr>$cost_url[string range $cost_name 0 30]</A></nobr></td>
	  <td>$cost_center_code</td>
	  <td>$company_name</td>
	  <td>$calculated_due_date</td>
	  <td><nobr>$amount_converted $default_currency_read_p</nobr></td>
	  <td><nobr>$amount_unconverted</td>
	  <td><nobr>$amount_paid</nobr></td>
	</tr>\n"
	incr ctr
    }

    # Write the subtotal line of the last cost_type_id section
    if {$ctr > 1} {
	if {!$atleast_one_unreadable_p} {
	    append cost_html "
		<tr class=rowplain>
		  <td colspan=[expr $colspan-3]>&nbsp;</td>
		  <td colspan=2>
		    <b>$subtotals($old_cost_type_id) $default_currency</b>
		  </td>
		</tr>
            "
	}
	append cost_html "
		<tr>
		  <td colspan=99 class=rowplain>&nbsp;</td>
		</tr>\n"
    }

    # Add a reasonable message if there are no documents
    if {$ctr == 1} {
	append cost_html "
	<tr$bgcolor([expr $ctr % 2])>
	  <td colspan=$colspan align=center>
	    <I>[_ intranet-cost.lt_No_financial_document]</I>
	  </td>
	</tr>\n"
	incr ctr
    }

    # Close the main table
    append cost_html "</table>\n"

    if {!$show_details_p} { set cost_html "" }


    # ----------------- Hard Costs HTML -------------
    # Hard "real" costs such as invoices, bills and timesheet

    set hard_cost_html "
<table with=\"100%\">
  <tr class=rowtitle>
    <td class=rowtitle colspan=2 align=center>[_ intranet-cost.Real_Costs]</td>
  </tr>
  <tr>
    <td>[_ intranet-cost.Customer_Invoices]</td>\n"
    set subtotal $subtotals([im_cost_type_invoice])
    append hard_cost_html "<td align=right>$subtotal $default_currency</td>\n"
    set grand_total $subtotal

    append hard_cost_html "</tr>\n<tr>\n<td>[_ intranet-cost.Provider_Bills]</td>\n"
    set subtotal $subtotals([im_cost_type_bill])
    append hard_cost_html "<td align=right>- $subtotal $default_currency</td>\n"
    set grand_total [expr $grand_total - $subtotal]

    append hard_cost_html "</tr>\n<tr>\n<td>[_ intranet-cost.Timesheet_Costs]</td>\n"
    set subtotal $subtotals([im_cost_type_timesheet])
    append hard_cost_html "<td align=right>- $subtotal $default_currency</td>\n"
    set grand_total [expr $grand_total - $subtotal]

    append hard_cost_html "</tr>\n<tr>\n<td>[lang::message::lookup "" intranet-cost.Expenses "Expenses"]</td>\n"
    set subtotal $subtotals([im_cost_type_expense_bundle])
    append hard_cost_html "<td align=right>- $subtotal $default_currency</td>\n"
    set grand_total [expr $grand_total - $subtotal]

    append hard_cost_html "</tr>\n<tr>\n<td><b>[_ intranet-cost.Grand_Total]</b></td>\n"
    append hard_cost_html "<td align=right><b>$grand_total $default_currency</b></td>\n"
    append hard_cost_html "</tr>\n</table>\n"


    # ----------------- Prelim Costs HTML -------------
    # Preliminary (planned) Costs such as Quotes and Purchase Orders

    set prelim_cost_html "
<table width=\"100%\">
  <tr class=rowtitle>
    <td class=rowtitle colspan=2 align=center>[_ intranet-cost.Preliminary_Costs]</td>
  </tr>
  <tr>
    <td>[_ intranet-cost.Quotes]</td>\n"

    set subtotal $subtotals([im_cost_type_quote])
    append prelim_cost_html "<td align=right>$subtotal $default_currency</td>\n"
    set grand_total $subtotal

    append prelim_cost_html "</tr>\n<tr>\n<td>[_ intranet-cost.Purchase_Orders]</td>\n"
    set subtotal $subtotals([im_cost_type_po])
    append prelim_cost_html "<td align=right>- $subtotal $default_currency</td>\n"
    set grand_total [expr $grand_total - $subtotal]

    append prelim_cost_html "</tr>\n<tr>\n<td>[lang::message::lookup "" intranet-cost.Timesheet_Planned_Costs "Timesheet Planned"]</td>\n"
    append prelim_cost_html "<td align=right>- $subtotals([im_cost_type_timesheet_planned]) $default_currency</td>\n"

    append prelim_cost_html "</tr>\n<tr>\n<td>[lang::message::lookup "" intranet-cost.Expenses_Planned_Costs "Expenses Planned"]</td>\n"
    append prelim_cost_html "<td align=right>
	  $subtotals([im_cost_type_expense_planned]) $default_currency
	</td>\n"

    append prelim_cost_html "</tr>\n<tr>\n<td><b>[lang::message::lookup "" intranet-cost.Preliminary_Total "Preliminary Total"]</b></td>\n"
    append prelim_cost_html "<td align=right><b>$grand_total $default_currency</b></td>\n"
    append prelim_cost_html "</tr>\n</table>\n"


    # ----------------- Check that the Exchange Rates are still valid  -------------

    # See which currencies are to be added here...
    set used_currencies [db_list currencies_used "
	select distinct
		ci.currency
	from	im_costs ci
	where	ci.currency != :default_currency
		and cost_id in (
		    $project_cost_ids_sql
		)
    "]

    set exchange_rates_outdated [im_exchange_rate_outdated_currencies]
    set currency_outdated_warning ""
    if {[llength $used_currencies] > 0 && [llength $exchange_rates_outdated] > 0} {

	set currency_outdated_warning [lang::message::lookup "" intranet-cost.The_exchange_rates_are_outdated "The exchanges rates for the following currencies are outdated. <br>Please contact your System Administrator to update the following exchange rates:"]
	append currency_outdated_warning "\n<br>\n"
	append currency_outdated_warning "\n<table>\n"

	foreach entry $exchange_rates_outdated {
	    set currency [lindex $entry 0]
	    set days [lindex $entry 1]
	    set outdated_msg [lang::message::lookup "" intranet-cost.Outdated_since_x_days "Outdated since %days% days"]
	    append currency_outdated_warning "\n<tr><td>$currency:</td><td>$outdated_msg</td></tr>\n"
	}
	append currency_outdated_warning "\n</table>\n"


	set currency_outdated_warning "
		<table>
		<tr class=rowtitle><td class=rowtitle align=center>
		[lang::message::lookup "" intranet-cost.Outdated_Message "Outdated Exchange Rates Warning"]
		</td></tr>
		<tr><td>$currency_outdated_warning</td></tr>
		</table>
        "

    }


    # ----------------- Admin Links --------------------------------

    # Restore value overwritten by SQL query
    set project_id $org_project_id
    
    # Add some links to create new financial documents
    # if the intranet-invoices module is installed
    set admin_html ""
    if {$show_admin_links_p && [db_table_exists im_invoices]} {

	set admin_html "
	<table>
	<tr>
	  <td class=rowtitle align=center>
	    [_ intranet-core.Admin_Links]
	  </td>
	</tr>
	<tr class=rowplain>
	  <td>\n"

	    # Customer invoices: customer = Project Customer, provider = Internal
	    set customer_id [db_string project_customer "select company_id from im_projects where project_id = :org_project_id" -default ""]
	    set provider_id [im_company_internal]
	    set bind_vars [ad_tcl_vars_to_ns_set project_id customer_id provider_id]
	    append admin_html [lang::message::lookup "" intranet-cost.Customer_Links "Customer Actions"]
      	    append admin_html [im_menu_ul_list "invoices_customers" $bind_vars]

	    # Provider invoices: customer = Internal, no provider yet defined
	    set customer_id [im_company_internal]
	    set bind_vars [ad_tcl_vars_to_ns_set customer_id project_id]
	    append admin_html [lang::message::lookup "" intranet-cost.Provider_Links "Provider Actions"]
	    append admin_html [im_menu_ul_list "invoices_providers" $bind_vars]

	    append admin_html "	
	  </td>
	</tr>
	</table>
	"
    }

    # ----------------- Assemble the "Summary" component ---------------------------
    # With preliminary and hard costs

    set summary_html ""
    if {$show_details_p} {

	# Summary in broad format
	set summary_html "
	<table cellspacing=0 cellpadding=0>
	<tr><td class=rowtitle colspan=3 align=center>[_ intranet-cost.Financial_Summary]</td></tr>
	<tr valign=top>
	  <td>$hard_cost_html</td>
	  <td>&nbsp &nbsp;</td>
	  <td>$prelim_cost_html</td>
	</tr>
	</table>\n"

    } else {

	# Summary in narrow format
	set summary_html "
	<table cellspacing=0 cellpadding=0 width=\"100%\">
	<tr><td class=rowtitle align=center>[_ intranet-cost.Financial_Summary]</td></tr>
	<tr valign=top><td>$hard_cost_html</td></tr>
	<tr><td>$prelim_cost_html</td></tr>
	</table>\n"

	set summary_html "
	$hard_cost_html
        <br>
	$prelim_cost_html
        "

    }

    if {!$show_summary_p} { set summary_html "" }
    if {!$can_read_summary_p} { set summary_html "" }

    # ----------------- Put the component/page together ---------------------------

    set result_html "
        $currency_outdated_warning
	<table>
	<tr valign=top>
	  <td>
	    $cost_html
            <br>
            $summary_html
	  </td>
	  <td>
	    $admin_html
	  </td>
	</tr>
        <tr><td colspan=2>
        </td></tr>
	</table>
    "

    return $result_html
}


ad_proc -public im_cost_type_select { 
    select_name 
    { default "" } 
    { super_type_id 0 } 
    { cost_item_group "" }
} {
    Returns an html select box named $select_name and defaulted to
    $default with a list of all the cost_types in the system.
    If super_type_id is specified then return only those types "below" super_type.

    @param cost_item_group Can be "" (all) or "financial_doc".
           Limits the number of cost types to a certain group of FinDocs
} {
    set category_type "Intranet Cost Type"
    set bind_vars [ns_set create]
    ns_set put $bind_vars category_type $category_type

    set sql "
	select	c.category_id,
		c.category
	from	im_categories c
	where	c.category_type = :category_type and
		(enabled_p is null OR enabled_p = 't')
    "

    # Restrict to specific subtypes of FinDocs
    switch [string tolower $cost_item_group] {
	"financial_doc" {
	    append sql "\n\t\tand c.category_id in (
		[im_cost_type_invoice],
		[im_cost_type_quote],
		[im_cost_type_bill],
		[im_cost_type_po],
		[im_cost_type_delivery_note]
		)"
	}
    }

    if {$super_type_id} {
	ns_set put $bind_vars super_type_id $super_type_id
	append sql "\n	and c.category_id in ([join [im_sub_categories $super_type_id] ","])"
    }
    return [im_selection_to_select_box $bind_vars category_select $sql $select_name $default]
}



ad_proc -public im_cost_status_select { {-translate_p 1}  select_name { default "" } } {
    Returns an html select box named $select_name and defaulted to
    $default with a list of all the cost status_types in the system
} {
    set include_empty 0
    set options [util_memoize "im_cost_status_options $include_empty"]

    set result "\n<select name=\"$select_name\">\n"
    if {[string equal $default ""]} {
	append result "<option value=\"\"> -- Please select -- </option>"
    }

    foreach option $options {

	if { $translate_p } {
	    set text [_ intranet-core.[lang::util::suggest_key [lindex $option 0]]]
	} else {
	    set text [lindex $option 0]
	}
	
	set selected ""
	if { [string equal $default [lindex $option 1]]} {
	    set selected " selected"
	}
	append result "\t<option value=\"[util_quote_double_quotes [lindex $option 1]]\" $selected>"
	append result "$text</option>\n"

    }

    append result "\n</select>\n"
    return $result
}


ad_proc im_cost_payment_method_select { select_name { default "" } } {
    Returns an html select box named $select_name and defaulted to $default 
    with a list of all the partner statuses in the system
} {
    return [im_category_select "Intranet Cost Payment Method" $select_name $default]
}

ad_proc im_cost_template_select { select_name { default "" } } {
    Returns an html select box named $select_name and defaulted to $default 
    with a list of all the partner statuses in the system
} {
    return [im_category_select_plain -translate_p 0 "Intranet Cost Template" $select_name $default]
}



ad_proc im_costs_select { select_name { default "" } { status "" } { exclude_status "" } } {
    Returns an html select box named $select_name and defaulted to
    $default with a list of all the costs in the system. If status is
    specified, we limit the select box to costs that match that
    status. If exclude status is provided, we limit to states that do not
    match exclude_status (list of statuses to exclude).

} {
    set bind_vars [ns_set create]

    set sql "
select
	i.cost_id,
	i.cost_name
from
	im_costs i
where
	1=1
"

    if { ![empty_string_p $status] } {
	ns_set put $bind_vars status $status
	append sql " and cost_status_id=(select cost_status_id from im_cost_status where cost_status=:status)"
    }

    if { ![empty_string_p $exclude_status] } {
	set exclude_string [im_append_list_to_ns_set $bind_vars cost_status_type $exclude_status]
	append sql " and cost_status_id in (select cost_status_id 
						from im_cost_status 
						where cost_status not in ($exclude_string)) "
    }
    append sql " order by lower(cost_name)"
    return [im_selection_to_select_box $bind_vars "cost_status_select" $sql $select_name $default]
}


ad_proc im_cost_update_payments { cost_id } {
    Update the payment amount for a cost item by
    summing up all payments for this item.
} {
    set default_currency [ad_parameter -package_id [im_package_cost_id] "DefaultCurrency" "" "EUR"]

    db_dml update_cost_payment "
	update im_costs set 
	    paid_amount = (
	        select  sum(amount * im_exchange_rate(received_date::date, currency, :default_currency))
	        from    im_payments
	        where   cost_id = :cost_id
	    ),
	    paid_currency = :default_currency
	where cost_id = :cost_id
    "
}





# -----------------------------------------------------------
# Cost Cache Sweeper
#
# This is an offline process that is scheduled every ~60 seconds
# to check for projects with the "cost_cache_dirty" NOT NULL.
# 
# Projects with this mark have suffered from changes in the 
# associated financial documents, possibly from sub-projects,
# or structural changes (new or removed subproject).
#
# So this sweeper updates the cost_cache (various "cache" fields
# of im_projects), keeping the controlling information up to date
# within the given limits.
# -----------------------------------------------------------

ad_proc -public im_cost_cache_sweeper { } {
    Checks for the MaxCount projects with the "oldest" outdated
    cost caches. The "im_cost_project_cache_invalidator" does not
    overwrite existing "cost_cache_dirty" entries, so that this
    fields contains the first modification dates in a correct order,
    allowing us an "oldest first" update. This way we best and
    deterministicly limit the average outdated time per cache.
} {
    set max_projects [parameter::get_from_package_key -package_key intranet-cost -parameter "CostCacheSweeperMaxProjects" -default 50]

    set sql "
	select	project_id
	from	im_projects
	where	cost_cache_dirty is not null
	order by cost_cache_dirty
	LIMIT :max_projects
    "
    db_foreach sweep_cost_cache_projects $sql {
	# Update the sum of financial items across subprojects
	im_cost_update_project_cost_cache $project_id
    }
}





ad_proc -public im_cost_update_project_cost_cache { 
    project_id 
} {
    Calculates the sums of all cost elements per project,
    including subprojects of arbitrary depth.
    Returns the "subtotals" array.
} {
    set default_currency [ad_parameter -package_id [im_package_cost_id] "DefaultCurrency" "" "EUR"]
    set default_hourly_cost [ad_parameter -package_id [im_package_cost_id] "DefaultHourlyCost" "" 30]

    # Update the logged hours cache
    im_timesheet_update_timesheet_cache -project_id $project_id

    set project_cost_ids_sql "
		                select distinct cost_id
		                from im_costs
		                where project_id in (
					select	children.project_id
					from	im_projects parent,
						im_projects children
					where	children.tree_sortkey 
							between parent.tree_sortkey 
							and tree_right(parent.tree_sortkey)
						and parent.project_id = :project_id
				)
			    UNION
				select distinct object_id_two as cost_id
				from acs_rels
				where object_id_one in (
					select	children.project_id
					from	im_projects parent,
						im_projects children
					where	children.tree_sortkey 
							between parent.tree_sortkey 
							and tree_right(parent.tree_sortkey)
						and parent.project_id = :project_id
				)
    "

   
    set subtotals_sql "
	select
		sum(ci.amount_converted) as amount_converted,
	        cat.cost_type_id
	from
		im_cost_types cat 
		LEFT OUTER JOIN	(
			select	ci.*,
				round((im_exchange_rate(
					ci.effective_date::date, 
					ci.currency, 
					:default_currency
				) * amount)::numeric, 2) as amount_converted
			from	im_costs ci
			where	
				ci.cost_id in (
					$project_cost_ids_sql
				)
		) ci on (cat.cost_type_id = ci.cost_type_id)
	where
		cat.cost_type_id not in (
			[im_cost_type_employee],
			[im_cost_type_repeating],
			[im_cost_type_expense_item]
		)
		and ci.currency is not null
	group by
		cat.cost_type_id
    "

    set cost_type_sql "select category_id from im_categories where category_type='Intranet Cost Type' and (enabled_p is null OR enabled_p = 't') "
    db_foreach subtotal_init $cost_type_sql {
	set subtotals($category_id) 0
    }

    db_foreach subtotals $subtotals_sql {
	if {"" == $amount_converted} { set amount_converted 0 }
        set subtotals($cost_type_id) $amount_converted
    }

    # Special treatment for timesheet hours budget - multiply with default hourly rate
    set budget_hours [db_string budget_hours "select project_budget_hours from im_projects where project_id = :project_id" -default ""]
    if {"" == $budget_hours} { set budget_hours 0 }
    set cost_timesheet_planned [expr $budget_hours * $default_hourly_cost]
    set subtotals([im_cost_type_timesheet_planned]) $cost_timesheet_planned

    # Expense Planned
    set subtotals([im_cost_type_expense_planned]) 0
    

    # We can update the profit & loss because all financial documents
    # have been converted to default_currency
    db_dml update_projects "
		update im_projects set
			cost_invoices_cache = $subtotals([im_cost_type_invoice]),
			cost_bills_cache = $subtotals([im_cost_type_bill]),
			cost_timesheet_logged_cache = $subtotals([im_cost_type_timesheet]),
			cost_expense_logged_cache = $subtotals([im_cost_type_expense_bundle]),
			cost_quotes_cache = $subtotals([im_cost_type_quote]),
			cost_purchase_orders_cache = $subtotals([im_cost_type_po]),
			cost_delivery_notes_cache = $subtotals([im_cost_type_delivery_note]),
			cost_timesheet_planned_cache = :cost_timesheet_planned,
			cost_expense_planned_cache = 0,
			cost_cache_dirty = null
		where
			project_id = :project_id
    "

    return [array get subtotals]
}



# -----------------------------------------------------------
# NavBar tree for finance
# -----------------------------------------------------------

ad_proc -public im_navbar_tree_finance { } { Finance Navbar } {
    return "
	<li><a href=/intranet/>[lang::message::lookup "" intranet-cost.Finance "Finance"]</a>
	<ul>
		<li><a href=/intranet-invoices/list?cost_type_id=3708>New Cust. Invoices &amp; Quotes</a>
		<ul>
			[im_navbar_write_tree -label "invoices_customers" -maxlevel 0]
		</ul>
		<li><a href=/intranet-invoices/list?cost_type_id=3710>New Prov. Bills &amp; POs</a>
		<ul>
			[im_navbar_write_tree -label "invoices_providers" -maxlevel 0]
		</ul>
		<li><a href=/intranet-invoices/list?cost_status_id=3802&cost_type_id=3700>Accounts Receivable</a></li>
		<li><a href=/intranet-invoices/list?cost_status_id=3802&cost_type_id=3704>Accounts Payable</a></li>
		<li><a href=/intranet-payments/index>Payments</a></li>
		<li><a href=/intranet-dw-light/invoices.csv>Export Finance to CSV/Excel</a></li>

		<li><a href=/intranet-reporting/>Reporting</a>
                <ul>
                [im_navbar_write_tree -label "reporting-finance" -maxlevel 1]
                [im_navbar_write_tree -label "reporting-timesheet" -maxlevel 1]
                </ul>

		<li><a href=/intranet/admin/>Admin</a>
		<ul>
			<li><a href=/intranet-cost/cost-centers/index>Cost Centers &amp Departments</a>
			<li><a href=/intranet-exchange-rate/index>Exchange Rates</a>
			<li><a href=/intranet-material/>Materials (Service Types)</a>
		</ul>
	</ul>
    "
}



